#!/usr/bin/env python

'''
client.py: simple clients for google storage. This first go uses
datastore for metadata, and Google Storage for images (objects) so
a client means a connection to both, with functions to interact with 
both. Both will look for the environment variable GOOGLE_APPLICATION_CREDENTIALS

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


'''

import datetime

from google.cloud import datastore
from som.api.google.storage.utils import (
    get_google_service, 
    get_bucket,
    upload_file
)

from som.api.google.storage.models import BatchManager
from som.api import ApiConnection
from som.logger import bot


class ClientBase(ApiConnection):

    def __init__(self,**kwargs):
        super(ApiConnection, self).__init__(**kwargs)
        self.datastore = datastore.Client(self.project_name)
        self.batch = BatchManager(client=self.datastore)
        self.storage = get_google_service('storage', 'v1')
        if self.bucket_name is not None:
            self.get_bucket()
  
    def get_bucket(self):
        self.bucket = get_bucket(self.storage,self.bucket_name)

    def make_key(self,key,ancestor=None):
        if ancestor is not None:
            ancestor = list(ancestor.key._flat_path)
            key = ancestor + key
        return self.datastore.key(*key)


    def put_object(self,bucket_folder,file_path,verbose=True,permission=None, mimetype=None):
        '''upload_object will upload a file to path bucket_path in storage
        '''
        return upload_file(storage_service=self.storage,
                           bucket=self.bucket,
                           mimetype=mimetype,
                           bucket_path=bucket_folder,
                           file_path=file_path,
                           permission=permission,
                           verbose=verbose)



'''
client.py: client for working with identifiers API

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from som.api.base import SomApiConnection
from .standards import (
    spec
)

from som.logger import bot
import sys
import os

api_base = "https://api.rit.stanford.edu/identifiers/api"
api_version = "v1"

class Client(SomApiConnection):

    def __init__(self, token=None, study=None):

        if study is None:
            study = "test"
        bot.info("Client: <study:%s>" %(study))
        self.base = api_base
        self.version = api_version
        self.study = study
        self.spec = spec
        self.token = token
        super(Client, self).__init__()

    def __str__(self):
        return "Client: <study:%s>" %(self.study)

    def __repr__(self):
        return "Client: <study:%s>" %(self.study)


    def deidentify(self,ids,study=None):
        '''deidentify: uid endpoint: 
         https://api.rit.stanford.edu/identifiers/api/v1/uid/{study} 
         will take a list of identifiers, and return the deidentified.        
        :param ids: a list of identifiers
        :param study: if None, defaults to test.
        '''    
        # Testing overrides all other specifications
        if study is None:
            study = self.study
        study = study.lower()
        bot.debug("study: %s" %study)

        url = "%s/%s/uid/%s" %(self.base,
                              self.version,
                              study)

        return self.post(url=url,data=ids)



'''
validators.py simple validation and cleaning of model data structures

'''

def validate_model(fields):
    '''remove_nones will remove any None fields from a passed dictionary
    '''
    keepers = dict()
    for entry in fields:
        if entry['value'] is not None:
            keepers[entry['key']] = entry['value']
        else:
            if entry['required'] == True:
                bot.logger.error("Field %s is required for this entity.",entry['key'])
    return keepers




#!/usr/bin/env python

'''

validators.py: validation of folders and compressed objects for wordfish standard

'''

from som.logger import bot
import tempfile
import shutil
import os
import re
import json
import sys

from som.utils import (
    untar_dir,
    unzip_dir,
    read_json
)

from validator import (
    Required, 
    Not, 
    Truthy, 
    Blank, 
    Range, 
    Equals, 
    In, 
    validate
)

def validate_dataset(dataset,testing_base=None,clean_up=True):
    '''validate_dataset is a general function to take a compressed object 
    (zip or .targz) or folder and run the correct validation functions depending
    on the input type.
    :param dataset: the full path to the dataset
    :param testing_base: the testing_base directory
    :param clean_up: whether to clean up the directory upon finish
    '''
    if os.path.isdir(dataset):
        valid = validate_folder(folder=dataset)
    elif re.search("[.]zip$|[.]tar[.]gz$",dataset):
        valid = validate_compressed(compressed_file=dataset,
                                    testing_base=testing_base,
                                    clean_up=clean_up)
    return valid


def validate_folder(folder):
    '''validate_folder will ensure the folder provided is reflective of WordFish standard
    :param folder: the folder to validate, corresponding to a collection base
    :returns valid: True/False
    '''    

    # The first level should have folders with collections
    valid = True
    full_path = os.path.abspath(folder)

    if validate_metadata(full_path) == False:        
        valid = False

    # validate images, text, and metadata of the entities
    valid_entities = validate_entities(full_path)
    if valid_entities == None:
        bot.error("found invalid collection: does not have entities.")
        valid = False        
    elif valid_entities == False:
        bot.error("found invalid collection: invalid entities.")
        valid = False        
    else:
        print("collection %s is valid." %folder)        

    return valid


def validate_metadata(full_path,metadata_type=None):
    '''validate_metadata checks to see if a name (either a collection
    name, folder for an image or text) has associated metadata, indicated by
    a file of the same name (ending with json) in the parent directory of the
    named file. If no matching files are found, None is returned, and the user
    is alerted. If a matching file is found, it is checked to be valid json.
    :param full_path: full path to a file or folder
    :param metadata_type: either one of collection, image, or text. Default collection
    '''
    if metadata_type == None:
        metadata_type = "collection"

    parent_dir = os.path.dirname(full_path)
    base_name = os.path.basename(full_path).split('.')[0]
    metadata = "%s/%s.json" %(parent_dir,base_name)

    if os.path.exists(metadata):
        bot.debug('found %s metadata: %s' %(metadata_type, base_name))
        try:
            md = read_json(metadata)
            bot.info('%s %s metadata is valid' %(metadata_type, base_name))
        except:
            bot.error('%s %s has invalid json metadata %s' %(metadata_type, base_name, metadata))
            return False
 
    else:
        bot.info('%s %s does not have metadata file %s.json' %(metadata_type, base_name, base_name))
        return None

    return True


def validate_entities(full_path):
    '''validate_entities will check to see if each subdirectory (an entity
    in a collection) has a set of valid images and text objects. The user
    is alerted about extraneous files.
    :param full_path: the full path to the collection folder with entities
    '''
    valid = True
    entities = os.listdir(full_path)
    bot.info("Found %s entities in collection." %len(entities))
    if len(entities) == 0:
        return None
     
    for entity in entities:
        entity_path = "%s/%s" %(full_path,entity)

        # Does the entity have metadata?
        if validate_metadata(entity_path,"entity") == False:      
            valid = False

        entity_texts = validate_texts(entity_path)
        entity_images = validate_images(entity_path)

        # If images and text are empty for a collection, invalid
        if entity_texts == None and entity_images == None:
            bot.error("found invalid entity: does not have images or text.")
            valid = False

        # if either text or images are not valid, entities considered invalid
        if entity_texts == False or entity_images == False:
            bot.error("entity %s does not have valid images or text." %(entity))
            valid = False

    return valid


def validate_texts(entity_path,acceptable_types=None):
    '''validate_texts will check an entity directory
    for a text folder. If it exists, each subfolder will be assessed for
    correct wordfish structure and metadata
    :param entity_path the path to the top level (entity) folder
    :param acceptable_types: the valid extensions to allow
    '''    
    if acceptable_types == None:
        acceptable_types = ['txt']

    valid = validate_template(entity_path=entity_path, 
                              template_type="text",
                              acceptable_types=acceptable_types)
    return valid



def validate_images(entity_path, acceptable_types=None):
    '''validate_images will check an entity directory
    for an images folder. If it exists, each subfolder will be assessed for
    correct wordfish structure and metadata
    :param entity_path the path to the top level (entity) folder
    :param acceptable_types: the valid extensions to allow
    '''
    
    if acceptable_types == None:
        acceptable_types = ['dcm','png','jpg','jpeg','nii','nii.gz']

    valid = validate_template(entity_path=entity_path, 
                              template_type="images",
                              acceptable_types=acceptable_types)
    return valid


def validate_template(entity_path, template_type, acceptable_types):
    '''validate_template will check an entity directory
    for an folder of a particular type, for files and metadata that
    meet a particular criteria. If needed, additional parsing
    functions can be passed to this function. 
    :param entity_path the path to the top level (entity) folder
    :param template_type: should be one of images or text
    :param acceptable_types: the valid extensions to allow
    '''
    valid = True
    template_path = "%s/%s" %(entity_path,template_type)
    entity_name = os.path.basename(entity_path)
    if not os.path.exists(template_path):
        bot.info("entity %s does not have %s." %(entity_name, template_type))
        return None
    
    # Let's keep track of each file
    all_folders = os.listdir(template_path)
    valids = []    # valid files
    others = []    # Not valid as metadata or accepted

    # Find all valid images
    for folder in all_folders:
        folder_path = "%s/%s" %(template_path,folder)
        all_files = os.listdir(folder_path)
        for single_file in all_files:
            file_path = "%s/%s" %(folder_path,single_file)
            parts = single_file.split('.')
            ext = '.'.join(parts[1:])
            if ext in acceptable_types:
                valids.append(file_path)
            else:
                others.append(file_path)

    # Warn the user about missing valid files, not logical given folder
    if len(valids) == 0:
        bot.warning("entity %s does not have %s." %(entity_name, template_type))
        return None
    else:
        bot.info("entity %s has %s %s" %(entity_name, len(valids), template_type))

    # Parse through the "others" and alert user about invalid file
    valid_metadata = 0
    invalid_metadata = 0
    skipped_files = 0    

    # Assess each valid for metadata
    for contender in valids:
        if validate_metadata(contender,template_type) == False:      
            bot.error("metadata %s for entity %s is invalid" %(contender,entity_name))
            invalid_metadata +=1
            valid = False
        else:
            valid_metadata +=1
    else:
        skipped_files +=1      
        bot.warning("%s for %s/%s is not valid for import and is ignored" %(contender, entity_name, template_type))

    bot.info("found %s valid metadata, %s invalid metadata, and %s skipped files for %s" %(valid_metadata,
                                                                                           invalid_metadata,
                                                                                           skipped_files,
                                                                                           entity_name))
    return valid


def validate_compressed(compressed_file,testing_base=None,clean_up=True):
    '''validate_compressed will first decompress a file to a temporary location,
    and then test if the folder is valid given the WordFish standard.
    :param compressed_file: the file to first extract.
    :param testing_base: If not given, a temporary location will be created. Otherwise,
    a folder will be made in testing_base.
    :param clean_up: clean up (remove) extracted files/folders after test. Default True
    '''
    if testing_base == None:
        testing_base = tempfile.mkdtemp()
 
    valid = True
    dest_dir = tempfile.mkdtemp(prefix="%s/" %testing_base)
    if compressed_file.endswith('.tar.gz'):
        test_folder = untar_dir(compressed_file,dest_dir)
 
    elif compressed_file.endswith('.zip'):
        test_folder = unzip_dir(compressed_file,dest_dir)

    else:
        bot.error("Invalid compressed file type: %s, exiting." %compressed_file)
        sys.exit(1)

    # Each object in the folder (a collection)
    collections = os.listdir(test_folder)
    bot.info("collections found: %s" %len(collections))
    for collection in collections:
        collection_path = "%s/%s" %(test_folder,collection)
        if validate_folder(collection_path) == False:
            bot.error("collection %s is invalid." %collection)

    if clean_up == True:
        shutil.rmtree(dest_dir)
    return valid



'''
google/datastore.py: general storage utility functions

'''

from google.cloud import datastore as ds
from google.cloud.datastore.key import Key
from som.logman import bot
import sys
import os


######################################################################################
# Datastore Utils
######################################################################################

def parse_keys(client,keys):
    '''parse_keys will take in a list of keys, either just single model key
    names, or lists with a model and specific identifier, and return
    client.key objects
    '''
    for key in keys:
        if isinstance(key,list):
            keys.append(client.key(key[0],key[1]))
        else:
            keys.append(client.key(key))
        return keys

def get_key_filters():
    return ['<','>','=','<=','>=']


def print_datastore_path(name):
    '''print_datastore path will print a complete uri for a datastore
    object, like Entity1:key1/Entity2/key2
    '''
    if isinstance(name,Key):
        name = list(name.flat_path)
    parts = [ name[i:i+2] for i in range(0, len(name), 2) ]
    return "/".join([":".join(x) for x in parts])


######################################################################################
# Testing/Retry Functions
######################################################################################

def stop_if_result_none(result):
    '''stop if result none will return True if we should not retry 
    when result is none, False otherwise using retrying python package
    '''
    do_retry = result is not None
    return do_retry






#!/usr/bin/env python

'''
auth.py: authentication functions for som api

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


from som.logger import bot

from som.utils import (
    read_json,
    write_json
)

import os
import json
import requests
import sys


def read_client_secrets():
    secrets = None
    token_file = os.environ.get("STANFORD_CLIENT_SECRETS",None)
    if token_file is not None:
        if os.path.exists(token_file):
            secrets = read_json(token_file)
    if secrets is None:
        bot.error('Cannot find STANFORD_CLIENT_SECRETS credential file path in environment.')
        sys.exit(1)
    return secrets


def authenticate():
    '''get_access_token will return the currently active access token
    from the client secrets file
    '''
    token = None
    secrets = read_client_secrets()
    if secrets is not None:
        token = secrets['accessToken']
    return token


def refresh_access_token():
    '''refresh access token reads in the client secrets from 
    the token file, and update the tokens, and save back to file
    '''
    token_file = os.environ.get("STANFORD_CLIENT_SECRETS",None)
    secrets = read_client_secrets()
    token = None

    # Query to update the token - must be on Stanford network
    if secrets is not None:
        response = requests.post(secrets['token_uri'],
                                 data=json.dumps({'refreshToken':secrets['refreshToken']}),
                                 headers={'Content-Type':"application/json"})

        if response.status_code == 200:

            response = response.json()
            secrets["accessToken"] = response['accessToken']
            secrets["refreshToken"] = response['refreshToken']
            print("Successfully refreshed access token.")
            token_file = write_json(secrets,token_file)
            token = secrets['accessToken']

    return token



'''

Copyright (c) 2017 Vanessa Sochat, All Rights Reserved

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from som.api.google.storage import Client
from som.utils import write_json
from som.logger import bot

from google.cloud.exceptions import Forbidden
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import storage
from retrying import retry

import tempfile
import os
import sys



try:
    from urllib.request import urlopen, Request
except ImportError:
    from urllib2 import urlopen, Request




def progress_download(collection_name,
                      output_folder,
                      suid,
                      project,
                      bucket_name,
                      query_entity=True,
                      filters=None):

    '''
    show progress while downloading images for a Collection/[c]/Entity/study 
    
    Parameters
    ==========

    collection_name: the name of the collection, typically an IRB number
    output_folder: the base directory to create a study folder in
    project: Google Cloud project name
    suid: an suid of interest to query (eg, if querying an Entity, you would
          use the suid of the patient, an Image would be an suid of the
          study SUID --> (coded accession#)
    query_entity: by default, we query the entity first, and then get images.
                  to query the images (studies) set this to False.
    bucket_name: the name for the Google Storage Bucket (usually provided)
    filters: a list of tuples to apply to filter the query. Default is:

         [ ("entity_id","=", study) ]

    to retrieve all Image items that are equal to the study name

    Returns
    =======
    path to newly created image file

    '''

    if filters is None:
        if query_entity is True:
            filters = [ ("uid","=", suid) ]
        else:
            filters = [ ("AccessionNumber","=", suid) ]

    bot.info("Collecting available images...")

    # Retrieve bucket, datastore client, images
    try:
        storage_client = storage.Client()

    except DefaultCredentialsError:
        bot.error("We didn't detect your GOOGLE_APPLICATION_CREDENTIALS in the environment! Did you export the path?")
        sys.exit(1)
    except Forbidden:
        bot.error("The service account specified by GOOGLE_APPLICATION_CREDENTIALS does not have permission to use this resource.")
        sys.exit(1)

    if not os.path.exists(output_folder):
        os.mkdir(output_folder)

    bucket = storage_client.get_bucket(bucket_name)
    client = retry_get_client(bucket_name,project)
    collection = retry_get_collection(client,collection_name)

    if query_entity is True:
        entity_set = retry_get_entity(client,filters)
        images = []
        for entity in entity_set:
            entity_images = client.get_images(entity=entity)
            images = [x for x in entity_images if x not in images]
    else:
        images = retry_get_images(client,filters)
    
    bot.info("Found %s images for suid %s in collection %s" %(len(images),
                                                             suid,
                                                             collection_name))
    
    progress = 0
    total = len(images)

    files = []
    if len(images) > 0:
        bot.debug("Saving images and metadata...")
        for image in images:

            # Download image
            file_name = prepare_folders(output_folder=output_folder,
                                        image_name = image.key.name)
            
            blob = bucket.blob(image['storage_name'])

            bot.show_progress(progress, total, length=35)
            download_retry(blob,file_name)
            files.append(file_name)
            files.append(save_metadata(image,file_name))
            progress+=1
            bot.show_progress(progress,total,length=35)

        # Newline to finish
        sys.stdout.write('\n')

    return files


     
       
def save_metadata(image,file_name):
    '''
    save the image metadata to json "file_name", removing the created
    and updated timestamps from storage, to not confuse the user
    
    Parameters
    ==========
    image: a Google DataStore Entity that converts to dictionary
    file_name: the file name of the dicom the metadata is for

    Returns
    =======
    path to newly created image file

    '''
    metadata = dict(image)
    del metadata['created']
    del metadata['updated'] 
    metadata_file = file_name.replace('.dcm','.json')
    return write_json(metadata,metadata_file)



def prepare_folders(output_folder,image_name):
    '''prepare download path for image, and return path.
    '''

    folders = image_name.split('/')[:-1] # last is the image
    image_name = image_name.split('/')[-1]
    for folder in folders:
        output_folder = "%s/%s" %(output_folder,folder)
        if not os.path.exists(output_folder):
            os.mkdir(output_folder)
 
    return "%s/%s" %(output_folder,image_name)



def download_collection(collection,
                        project,
                        suid,
                        bucket,
                        query_entity=True,
                        output_folder=None,
                        filters=None):

    '''
    client function to download a collection in entirety, intended for
    command line application som get

    See parameters in progress_download
    '''

    if output_folder is None:
        output_folder = os.path.getcwd()

    if not os.path.exists(output_folder):
        bot.error("Output folder %s not found. Exiting." %output_folder)
        sys.exit(1)

    output_folder = "%s/%s" %(output_folder,collection)
    return progress_download(output_folder=output_folder,
                             suid=suid,
                             collection_name=collection,
                             project=project,
                             query_entity=query_entity,
                             bucket_name=bucket,
                             filters=filters)



########################################################################
# RETRY FUNCTIONS
#----------------------------------------------------------------------
# exponential backoff for Google Cloud APIs that can have hiccups from
########################################################################


@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000,stop_max_attempt_number=5)
def retry_get_collection(client,collection_name):
    return client.create_collection(collection_name)


@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000,stop_max_attempt_number=5)
def retry_get_entity(client,filters):
    return client.batch.query(kind="Entity",
                              filters=filters)


@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000,stop_max_attempt_number=5)
def retry_get_images(client,filters):
    return client.batch.query(kind="Image",
                              filters=filters)

@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000,stop_max_attempt_number=5)
def retry_get_client(bucket_name,project):
    return Client(bucket_name=bucket_name,
                  project_name=project)

@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000,stop_max_attempt_number=5)
def download_retry(blob,file_name):
    blob.download_to_filename(file_name)



#!/usr/bin/env python

'''
wordfish.base.py: base module for working with wordfish standard

'''

from som.logger import bot

from som.utils import (
    get_installdir, 
    read_file, 
    write_file
)

from som.wordfish.validators import (
    validate_dataset
)

from som.wordfish.structures import (
    structure_dataset
)

import subprocess
import tempfile
import zipfile
import inspect
import shutil
import requests
import imp
import sys
import re
import os


def run_validation(inputs,test_dir=None,clean_up=True,fail_exit=True):
    '''run validation will run one or more inputs through the validation procedure,
    meaning checking that the folder (and other data structures) fit the WordFish
    standard: http://www.github.com/radinformatics/wordfish-standard
    :param inputs: a single, or list of inputs, meaning folders and compressed files 
    for validation
    :param test_dir: a directory to use to extract and run things. If not specified,
    one is created.
    :param clean_up: boolean to determine if test_dir and subcontents should be
    removed after tests. Default is True.
    :param fail_exit: Given failure of validation, fail the process. Otherwise, return
    False to the calling function. Default fail_exit is True
    '''
    if not isinstance(inputs,list):
        inputs = [inputs]

    bot.debug("Found %s inputs to test using som-validator." %len(inputs))
    
    # Where are we testing?
    if test_dir == None:
        test_dir = tempfile.mkdtemp()

    # Tell the user about testing folder
    message = "Testing folder will be %s"
    if clean_up == True:
        message = "%s, and will be removed upon completion." %(message)
    bot.debug(message %test_dir)

    for testing in inputs:
        if os.path.isdir(testing):
            valid = validate_folder(folder=testing)
        elif re.search("[.]zip$|[.]tar[.]gz$",testing):
            valid = validate_compressed(compressed_file=testing,
                                        testing_base=test_dir,
                                        clean_up=clean_up)

        # Always exit or return False if input is not valid
        if valid == False:
            if fail_exit == True:
                bot.error("Input %s is not valid, please fix and retest. Exiting." %testing)
                sys.exit(1)
            bot.error("Input %s is not valid, please fix and retest. Returning False." %testing)
            return valid

    return valid


def get_structures(inputs,build_dir=None,clean_up=True,fail_exit=True):
    '''get structures will parse one or more compressed files and/ or folder paths
    and return a data structure that has full file paths for images/text documents,
    and the loaded json for metadata.
    :param inputs: a single, or list of inputs, meaning folders and compressed files 
    for validation
    :param build_dir: a directory to use to extract and run things. If not specified,
    one is created.
    :param clean_up: boolean to determine if test_dir and subcontents should be
    removed after tests. Default is True.
    :param fail_exit: Given failure of validation, fail the process. Otherwise, return
    False to the calling function. Default fail_exit is True
    '''
    if not isinstance(inputs,list):
        inputs = [inputs]

    bot.debug("Found %s inputs to structure using som-validator." %len(inputs))
    
    # Where are we testing?
    if build_dir == None:
        build_dir = tempfile.mkdtemp()

    # We will return a list of structures, each a collection
    structures = dict()

    # Tell the user about testing folder
    message = "Building folder will be %s"
    if clean_up == True:
        message = "%s, and will be removed upon completion." %(message)
    bot.debug(message %build_dir)

    for testing in inputs:
        valid = validate_dataset(dataset=testing,
                                 testing_base=build_dir,
                                 clean_up=clean_up)

        # We only structure input that is valid
        if valid == False:
            if fail_exit == True:
                bot.error("Input %s is not valid, please fix. Exiting." %testing)
                sys.exit(1)
            bot.error("Input %s is not valid, skipping.")
        else:
            structures[dataset] = structure_dataset(dataset=testing,
                                                    testing_base=build_dir,
                                                    clean_up=clean_up)

    return structures



'''
som.base: the base connection to the SOM API

'''

from som.logger import bot

from som.api.base.auth import (
    authenticate,
    refresh_access_token
)

from som.api import ApiConnection
from simplejson import JSONDecodeError as SimpleJSONDecodeError
from json import JSONDecodeError
import json
import sys

class SomApiConnection(ApiConnection):

    def __init__(self, **kwargs):
        super(SomApiConnection, self).__init__(**kwargs)
 
        self.headers = None
        if self.token is None:
            self.token = authenticate()
        self.update_headers()
        

    def call(self,url,func,data=None,return_json=True):
        '''call overrides post for the som api to add the option
        to refresh a token given a 401 response.
        :param func: the function (eg, post, get) to call
        :param url: the url to send file to
        :param data: additional data to add to the request
        :param return_json: return json if successful
        '''
        if isinstance(data,dict):
            data = json.dumps(data)

        response = func(url=url,
                        headers=self.headers,
                        data=data)

        # Errored response, try again with refresh
        if response.status_code == 401:
            bot.warning("Expired token, refreshing...")
            self.token = refresh_access_token()
            self.update_headers()
            response = func(url,
                            headers=self.headers,
                            data=data)

        if response.status_code == 200:

            if return_json:
                try:
                    response =  response.json()

                except (SimpleJSONDecodeError, JSONDecodeError):
                    bot.error("The server returned a malformed response. Are you on VPN?")
                    sys.exit(1)

        return response



#!/usr/bin/env python

'''
validators/responses.py: validation of json responses from api

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


'''

from som.logger import bot
import os
import sys

from validator import (
    Required, 
    Pattern,
    Not, 
    Truthy, 
    Blank, 
    Range, 
    Equals, 
    In, 
    validate
)

from som.api.identifiers.standards import (
    identifier_sources,
    item_sources,
    timestamp
)

from .utils import validate_fields


def receive_identifiers(response):
    '''receive identifiers will validate reception of an identifiers response.
    This should be a list
    :param response: the response list of identifiers

    :: notes
     successful response:

     HTTP 200

    [
       {'jittered_timestamp': '2016-01-30T17:15:23.123Z', 
        'id': '12345678', 
        'suid': '103e', 
        'id_source': 'Stanford MRN', 
        'custom_fields': [
             {'key': 'studySiteID', 'value': '78329'}], 
        'items': [
                   {
                    'id_source': 'GE PACS', 
                    'jittered_timestamp': '2016-01-15T17:15:23.123Z', 
                    'id': 'A654321', 
                    'suid': '103e'}
                   ]}

    ]

    '''
    # These fields are expected, but not required. We will error
    # if any fields are present outside this scope
    expected_fields = ['items',
                       'id_source', 
                       'jittered_timestamp', 
                       'suid', 
                       'id', 
                       'custom_fields']

    if not isinstance(response,list):
        bot.error("Response must be a list")
        return False

    # These are the rules for each uidEntity
    rules = {
        "id": [Required, Pattern("^[A-Za-z0-9_-]*$")], # pattern
        "suid": [Required, Pattern("^[A-Za-z0-9_-]*$")], # the suid
        "id_source": [Required, In(identifier_sources)],   # must be in identifer sources
        "jittered_timestamp": [Required,Pattern(timestamp)]
    }

    for item in response:

        # Validate required fields
        valid,message = validate(rules, item)
        if valid == False:
            bot.error(message)
            return valid

        # Validate fields returned in response
        if not validate_fields(expected_fields,item.keys()):
            return False

        # Validate items
        if "items" in item:
            if not receive_items(item['items']):
                return False
            
    bot.debug("Identifiers data structure valid: %s" %valid)
    return valid


def receive_items(items):
    '''receive items will validate reception of an items list.
    :param items: the items list from a response

     HTTP 200

        'items': [
                   {
                    'id_source': 'GE PACS', 
                    'jittered_timestamp': '2016-01-15T17:15:23.123Z', 
                    'id': 'A654321', 
                    'suid': '103e'
                    }
         ]

    '''
    expected_fields = ['id_source', 
                       'jittered_timestamp', 
                       'suid', 
                       'id',
                       'custom_fields']

    if not isinstance(items,list):
        bot.error("Items must be a list")
        return False

    # These are the rules for each uidEntity
    rules = {
        "id": [Required, Pattern("^[A-Za-z0-9_-]*$")], # pattern
        "suid": [Required, Pattern("^[A-Za-z0-9_-]*$")], # the suid
        "id_source": [Required, In(item_sources)],   # must be in person sources
        "jittered_timestamp": [Required,Pattern(timestamp)]
    }

    for item in items:

        # Validate required fields
        valid,message = validate(rules, item)
        if valid == False:
            bot.error(message)
            return valid

        # Validate fields returned in response
        if not validate_fields(expected_fields,item.keys()):
            return False

    return valid



#!/usr/bin/env python

'''
validators/utils.py: utility functions for validators

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


from som.logger import bot
import os
import sys

def validate_fields(acceptable_fields,actual_fields):
    '''validate_fields checks that there are not any in actual_fields
    that are not in acceptable_fields
    '''
    checking = [x for x in actual_fields if x not in acceptable_fields]
    if len(checking) == 0:
        return True
    bot.error("Fields not allowed: %s" %", ".join(checking))
    return False


def get_universal_source(source,comparator):
    '''get universal source will, given a source item (a list or single item), check to
    see if the length of a comparator is equivalent. The comparator is expected to be a list
    the same length as the source.
    :param source: the source item (single item or list)
    :param comparator: the thing to compare to
    :returns: EITHER a universal source (meaning one item to use for all items in source) OR None,
    meaning the comparator should be treated as a list, and source[i] matched to comparator[i]
    '''
    universal_source = None
    if isinstance(source,list):
        if len(source) != len(comparator):
            bot.error("Mismatch in length of source (%s) and comparator (%s). Exiting"
                                                            %(len(source),len(comparator)))
            sys.exit(1)
    else:
        # Otherwise, we assume a common source
        universal_source = id_sources
    return universal_source





from flask import (
    Flask, 
    render_template, 
    request
)

from som.wordfish.structures import structure_folder

from flask_restful import Resource, Api
from werkzeug import secure_filename
import webbrowser
import tempfile
import shutil
import random
import os

# SERVER CONFIGURATION ##############################################
class SomServer(Flask):

    def __init__(self, *args, **kwargs):
        super(SomServer, self).__init__(*args, **kwargs)

        # Set up temporary directory on start of application
        self.tmpdir = tempfile.mkdtemp()
        self.folder = None  # Dataset folder
        self.dataset = {} # Dataset json


# API VIEWS #########################################################
# eventually we can have views or functions served in this way...

app = SomServer(__name__)
api = Api(app)

class apiIndex(Resource):
    '''apiIndex
    Main view to show the entire dataset
    '''
    def get(self):
        return app.dataset


#class apiViewImage(Resource):
#    '''apiViewImage
#    shows an image 
#    '''
#    def get(self, url):
#        return {"ids": subset.id.tolist(),
#                "words":difference.index.tolist()}
              
# Add all resources
#api.add_resource(apiIndex,'/api/dataset') # start month, day, year / end month, date, year
#api.add_resource(apiViewImage,'/api/dataset/image/<path:url>')


@app.route('/som/annotate')
def annotate_images():
    return render_template('annotate_images.html')

@app.route('/som/dataset/explore')
def dataset_explorer():
    app.folder = 'cookies'
    app.dataset = structure_folder(app.folder,relative_path=True)
    collection_name = app.dataset['collection']['name']
    entities = app.dataset['collection']['entities']
    return render_template('dataset_explorer.html',entities=entities,
                                                   collection_name=collection_name)



# START FUNCTIONS ##################################################
    
# View a dataset folder
def run_dataset_explorer(folder,port=None):
    app.dataset = structure_folder(folder,relative_path=True)
    if port==None:
        port=8088
    print("It goes without saying. I suspect now it's not going.")
    webbrowser.open("http://localhost:%s/som/dataset/explorer" %(port))
    app.run(host="0.0.0.0",debug=False,port=port)


if __name__ == '__main__':
    app.debug = True
    app.run(host='0.0.0.0')



'''
google/datastore.py: general storage utility functions

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from google.cloud import datastore as ds
from google.cloud.datastore.key import Key
from som.logger import bot
import sys
import os


######################################################################################
# Datastore Utils
######################################################################################

def parse_keys(client,keys):
    '''parse_keys will take in a list of keys, either just single model key
    names, or lists with a model and specific identifier, and return
    client.key objects
    '''
    for key in keys:
        if isinstance(key,list):
            keys.append(client.key(key[0],key[1]))
        else:
            keys.append(client.key(key))
        return keys


def get_key_filters():
    return ['<','>','=','<=','>=']


def print_datastore_path(name):
    '''print_datastore path will print a complete uri for a datastore
    object, like Entity1:key1/Entity2/key2
    '''
    if isinstance(name,Key):
        name = list(name.flat_path)
    parts = [ name[i:i+2] for i in range(0, len(name), 2) ]
    return "/".join([":".join(x) for x in parts])


######################################################################################
# Testing/Retry Functions
######################################################################################

def stop_if_result_none(result):
    '''stop if result none will return True if we should not retry 
    when result is none, False otherwise using retrying python package
    '''
    do_retry = result is not None
    return do_retry






'''
client.py: simple clients for som api

'''

from som.api.base.connect import (
    api_base,
    api_version,
    SomApiConnection
)

from som.api.base.standards import spec
from som.logman import bot

class ClientBase(object):


    def __init__(self, spec=None, token=None, base=None, version=None):
 
        # Base and version consistent across
        if base == None:
            base = api_base

        if version == None:
            version = api_version

        self.base = base
        self.version = version
        self.study = None

        if spec != None:
            self.spec = spec

        self.api = SomApiConnection(token)


    def get_base(self,study=None):
        '''get_base returns the base api for a particular study
        '''
        if study is None:
            study = self.study
        return "%s/%s/%s" %(self.base,self.version,study)


    def deidentifiy_update(self,identifiers,test=False):
        '''deidentify update calls deidentify, but setting save_records
        to True to ensure that data is updated/written.
        '''
        return self.deidentify(identifiers,test=test,save_records=True)


    def deidentify(self,ids,test=False,save_records=False):
        '''deidentify will take a list of identifiers, and return the deidentified.
        if save_records is True, will save records. If False (default) only
        returns identifiers.
        :param identifiers: a list of identifiers (e.g., people)
        '''    
        # Default is production endpoint particular to the study
        study = self.study

        # Saving records (uid) or not (mrn) changes the endpoint
        action = "mrn"
        if save_records == True:
            bot.logger.debug("save_records is %s, no new data will be saved.")
            action = "uid"
        else:
            bot.logger.debug("save_records is %s, data will be saved.")

        # Testing overrides all other specifications
        if test == True:
            study = 'test'

        url = "%s/%s/%s/%s" %(api_base,api_version,action,study)

        response = self.api.post(url,data=ids)
        if "results" in response:
            return response['results']
        return response



#!/usr/bin/env python

'''

Copyright (c) 2017 Vanessa Sochat, All Rights Reserved

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import argparse
import sys
import os

def get_parser():
    parser = argparse.ArgumentParser(description="Stanford Open Modules for Python [SOM]")


    # Global Variables
    parser.add_argument("--version", dest='version', 
                        help="show software version", 
                        default=False, action='store_true')


    parser.add_argument('--debug', dest="debug", 
                        help="use verbose logging to debug.", 
                        default=False, action='store_true')


    subparsers = parser.add_subparsers(help='google storage and datastore',
                                       title='actions',
                                       description='actions for som tools',
                                       dest="command")


    # Get (download)
    get = subparsers.add_parser("get", 
                                help="download data from storge and datastore")

    get.add_argument("--outfolder", dest='outfolder', 
                     help="full path to folder for output, stays in tmp (or pwd) if not specified", 
                     type=str, default=None)

    get.add_argument('--project', dest="project",
                     help="name of Google Cloud project (eg, irlhs-learning)",
                     type=str, required=True)

    get.add_argument('--suid', dest="suid",
                     help='''An suid associated with an entity to find images (default) 
                             or query single images (set flag --query-images for the study. 
                             Typically begins with IRXXXX''',
                     type=str, required=True)

    get.add_argument('--query-images', dest="query_images", 
                        help="use verbose logging to debug.", 
                        default=False, action='store_true')

    get.add_argument('--bucket', dest="bucket",
                     help="Name of the storage bucket. Eg (irlhs-dicom)",
                     type=str, required=True)

    get.add_argument("--collection", dest='collection', 
                     help="name of collection (eg, IRB33192)", 
                     type=str, required=True)
    

    return parser


def get_subparsers(parser):
    '''get_subparser will get a dictionary of subparsers, to help with printing help
    '''
    actions = [action for action in parser._actions 
               if isinstance(action, argparse._SubParsersAction)]

    subparsers = dict()
    for action in actions:
        # get all subparsers and print help
        for choice, subparser in action.choices.items():
            subparsers[choice] = subparser

    return subparsers



def main():

    parser = get_parser()
    subparsers = get_subparsers(parser)

    try:
        args = parser.parse_args()
    except:
        sys.exit(0)

    # if environment logging variable not set, make silent
    if args.debug is False:
        os.environ['MESSAGELEVEL'] = "5"

    if args.version is True:
        from som.version import __version__
        print(__version__)
        sys.exit(0)


    if args.command == "get":
        from .get import download_collection 
        output_folder = download_collection(output_folder=args.outfolder,
                                            collection=args.collection,
                                            project=args.project,
                                            suid=args.suid,
                                            query_entity=not args.query_images,
                                            bucket=args.bucket)
        sys.exit(0)

    parser.print_help()

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import sys
import os

##########################################################################################
# HELPER FUNCTIONS #######################################################################
##########################################################################################

def get_lookup():
    '''get version by way of singularity.version, returns a 
    lookup dictionary with several global variables without
    needing to import singularity
    '''
    lookup = dict()
    version_file = os.path.join('som', 'version.py')
    with open(version_file) as filey:
        exec(filey.read(), lookup)
    return lookup


# Read in requirements
def get_requirements(lookup=None):
    '''get_requirements reads in requirements and versions from
    the lookup obtained with get_lookup'''

    if lookup == None:
        lookup = get_lookup()

    install_requires = []
    for module in lookup['INSTALL_REQUIRES']:
        module_name = module[0]
        module_meta = module[1]
        if "exact_version" in module_meta:
            dependency = "%s==%s" %(module_name,module_meta['exact_version'])
        elif "min_version" in module_meta:
            if module_meta['min_version'] == None:
                dependency = module_name
            else:
                dependency = "%s>=%s" %(module_name,module_meta['min_version'])
        install_requires.append(dependency)
    return install_requires


# Make sure everything is relative to setup.py
install_path = os.path.dirname(os.path.abspath(__file__)) 
os.chdir(install_path)

# Get version information from the lookup
lookup = get_lookup()
VERSION = lookup['__version__']
NAME = lookup['NAME']
AUTHOR = lookup['AUTHOR']
AUTHOR_EMAIL = lookup['AUTHOR_EMAIL']
PACKAGE_URL = lookup['PACKAGE_URL']
KEYWORDS = lookup['KEYWORDS']
DESCRIPTION = lookup['DESCRIPTION']
LICENSE = lookup['LICENSE']
with open('README.md') as filey:
    LONG_DESCRIPTION = filey.read()

##########################################################################################
# MAIN ###################################################################################
##########################################################################################


INSTALL_REQUIRES = get_requirements(lookup)

setup(
    name=NAME,
    version=VERSION,
    license=LICENSE,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=PACKAGE_URL,
    packages=find_packages(), 
    include_package_data=True,
    long_description=LONG_DESCRIPTION,
    keywords=KEYWORDS,
    install_requires=INSTALL_REQUIRES,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Shells',
        'Topic :: Terminals',
        'Topic :: Utilities'
    ],

    entry_points = {'console_scripts': [ 'som=som.app:main',
                                         'som-validator=som.wordfish.cli:main' ] }
)



#!/usr/bin/env python

'''
utils.py: part of som package

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import collections
import fnmatch
import os
import json
import re
import requests

import shutil
import simplejson
import som.__init__ as hello
from som.logger import bot
import sys

import subprocess

import tempfile
import tarfile
import zipfile


# Python less than version 3 must import OSError
if sys.version_info[0] < 3:
    from exceptions import OSError


######################################################################################
# Local commands and requests
######################################################################################

def get_installdir():
    '''get_installdir returns the installation directory of the application
    '''
    return os.path.abspath(os.path.dirname(hello.__file__))


def get_dataset(dataset=None):
    '''get_dataset will return some data provided by the application,
    based on a user-provided label. In the future, we can add https endpoints
    to retrieve online datasets.
    '''
    here = get_installdir()
    valid_datasets = {'developers_uid':'%s/api/identifiers/data/developers_uid.json' %here}
    if dataset is not None:

        # In case the user gave an extension
        dataset = os.path.splitext(dataset)[0].lower()
        if dataset in valid_datasets:
            return valid_datasets[dataset]

    bot.info("Valid datasets include: %s" %(','.join(list(valid_datasets.keys()))))


def run_command(cmd,error_message=None,sudopw=None,suppress=False):
    '''run_command uses subprocess to send a command to the terminal.
    :param cmd: the command to send, should be a list for subprocess
    :param error_message: the error message to give to user if fails, 
    if none specified, will alert that command failed.
    :param execute: if True, will add `` around command (default is False)
    :param sudopw: if specified (not None) command will be run asking for sudo
    '''
    if sudopw == None:
        sudopw = os.environ.get('pancakes',None)

    if sudopw != None:
        cmd = ' '.join(["echo", sudopw,"|","sudo","-S"] + cmd)
        if suppress == False:
            output = os.popen(cmd).read().strip('\n')
        else:
            output = cmd
            os.system(cmd)
    else:
        try:
            process = subprocess.Popen(cmd,stdout=subprocess.PIPE)
            output, err = process.communicate()
        except OSError as error: 
            if error.errno == os.errno.ENOENT:
                bot.error(error_message)
            else:
                bot.error(err)
            return None
    
    return output


def get_listfirst(item,group):
    '''return the first found in a list (group) from a 
    dictionary item.
    '''
    if not isinstance(group,list):
        group = [group]
    for contender in group:
        if contender in item:
            return item[contender]
    return None


############################################################################
## FILE OPERATIONS #########################################################
############################################################################

def write_file(filename,content,mode="w"):
    '''write_file will open a file, "filename" and write content, "content"
    and properly close the file
    '''
    with open(filename,mode) as filey:
        filey.writelines(content)
    return filename


def write_json(json_obj,filename,mode="w",print_pretty=True):
    '''write_json will (optionally,pretty print) a json object to file
    :param json_obj: the dict to print to json
    :param filename: the output file to write to
    :param pretty_print: if True, will use nicer formatting   
    '''
    with open(filename,mode) as filey:
        if print_pretty == True:
            filey.writelines(simplejson.dumps(json_obj, indent=4, separators=(',', ': ')))
        else:
            filey.writelines(simplejson.dumps(json_obj))
    return filename


def read_file(filename,mode="r"):
    '''write_file will open a file, "filename" and write content, "content"
    and properly close the file
    '''
    with open(filename,mode) as filey:
        content = filey.readlines()
    return content



def read_json(filename,mode="r"):
    '''read_json will open a file, "filename" and read the string as json
    '''
    with open(filename,mode) as filey:
        content = json.loads(filey.read())
    return content


############################################################################
## COMPRESSED FILES ########################################################
############################################################################


def detect_compressed(folder,compressed_types=None):
    '''detect compressed will return a list of files in 
    some folder that are compressed, by default this means
    .zip or .tar.gz, but the called can specify a custom list
    :param folder: the folder base to use.
    :param compressed_types: a list of types to include, should
    be extensions in format like *.tar.gz, *.zip, etc.
    '''
    compressed = []
    if compressed_types == None:
        compressed_types = ["*.tar.gz",'*zip']
    bot.debug("Searching for %s" %", ".join(compressed_types))

    for filey in os.listdir(folder):
        for compressed_type in compressed_types:
            if fnmatch.fnmatch(filey, compressed_type):
                compressed.append("%s/%s" %(folder,filey))
    bot.debug("Found %s compressed files in %s" %len(compressed),folder)
    return compressed


def unzip_dir(zip_file,dest_dir=None):
    '''unzip_dir will extract a zipfile to a directory. If
    an extraction destination is not defined, a temporary
    directory will be created and used.
    :param zip_file: the .zip file to unzip
    :param dest_dir: the destination directory
    '''

    if dest_dir == None:
        dest_dir = tempfile.mkdtemp()

    with zipfile.ZipFile(zip_file,"r") as zf:
        zf.extractall(dest_dir)
    return dest_dir



def zip_dir(zip_dir, zip_name, output_folder=None):
    '''zip_dir will zip up and entire zip directory
    :param folder_path: the folder to zip up
    :param zip_name: the name of the zip to return
    :output_folder:
    '''
    tmpdir = tempfile.mkdtemp()
    output_zip = "%s/%s" %(tmpdir,zip_name)
    zf = zipfile.ZipFile(output_zip, "w", zipfile.ZIP_DEFLATED, allowZip64=True)
    for root, dirs, files in os.walk(zip_dir):
        for file in files:
            zf.write(os.path.join(root, file))
    zf.close()
    if output_folder != None:
        shutil.copyfile(output_zip,"%s/%s"%(output_folder,zip_name))
        shutil.rmtree(tmpdir)
        output_zip = "%s/%s"%(output_folder,zip_name)
    return output_zip



def untar_dir(tar_file,dest_dir=None):
    '''untar_dir will extract a tarfile to a directory. If
    an extraction destination is not defined, a temporary
    directory will be created and used.
    :param tar_file: the .tar.gz file to untar/decompress
    :param dest_dir: the destination directory
    '''

    if dest_dir == None:
        dest_dir = tempfile.mkdtemp()

    contents = []
    if tarfile.is_tarfile(tar_file):
        with tarfile.open(tar_file) as tf:
            tf.extractall(dest_dir)
    return dest_dir



'''
tasks.py: functions to extract identifiers from dicom images

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


from som.logger import bot
from som.utils import (
    get_listfirst,
    read_json
)

import dateutil.parser
import tempfile

from deid.identifiers import (
    get_timestamp
)

from pydicom.sequence import Sequence

from som.api.identifiers.dicom.settings import (
   entity as entity_options,
   item as item_options
)

import os
import sys


######################################################################
# MAIN GET FUNCTIONS
######################################################################

def prepare_identifiers_request(ids, force=True):
    '''prepare identifiers request takes in the output from deid 
    get_identifiers, and returns the minimal request to send to DASHER.
    The assumption is that one batch of ids ==> one entity / Accession Number
    '''
    
    # Enforce application default
    entity_source = entity_options['id_source']['name']
    entity_field = entity_options['id_source']['field']
    item_source = item_options['id_source']['name']            
    item_field = item_options['id_source']['field']  
    entity_times = entity_options['id_timestamp']
    item_times = item_options['id_timestamp']

    # Entity --> Patient
    entity = {"id_source":entity_source,
              "items":[]}

    # Item --> Study (to represent all images)
    new_item = {"id_source": item_source}
    for item_id,item in ids.items():

        # Entity ID
        if "id" not in entity:
            entity_id = item[entity_field].replace('-','')
            entity["id"] = entity_id

        # Entity timestamp
        if "id_timestamp" not in entity:
            entity_ts = get_listfirst(item=item,group=entity_times['date']) # 20021202
            if entity_ts is not None: 
                timestamp = get_timestamp(item_date=entity_ts)              # 2002-12-02T00:00:00Z
                entity['id_timestamp'] = timestamp

        # Study Timestamp
        if "id_timestamp" not in new_item:
            item_ts = get_listfirst(item=item,group=item_times['date']) 
            if item_ts is not None:
                timestamp = get_timestamp(item_date=item_ts)
                new_item["id_timestamp"] = timestamp

        # Study ID (accession#)
        if "id" not in new_item:
            new_item["id"] = item[item_field]                                        

    # We are only including one study item to represent all images
    entity["items"].append(new_item)

    # Expected format of dasher
    ids = {"identifiers": [entity]}    
    return ids




def prepare_identifiers(response,ids,force=True):
    '''prepare identifiers is the step before replacing identifiers,
    it expects a response from the DASHER API, and will use it to prepare
    custom variables to each item provided in ids. It is assumed that the
    response given is relevant for all items in ids
    : param extracted: a dictionary with key (item id) and value a dictionary
    of extracted fields 
    '''

    # Format data correctly for deid
    for entity in response:      
        item = entity['items'][0]  
        updates = {'item_timestamp': item['jittered_timestamp'],
                   'entity_timestamp': entity['jittered_timestamp'],
                   'entity_id': entity['suid'],
                   'entity_id_source': entity['id_source'],
                   'item_id': item['suid'],
                   'item_id_source': item['id_source'],
                   'jitter': item['jitter'] }

    updated = dict()
    for item_id in ids:
        updated[item_id] = dict()
        updated[item_id].update(ids[item_id])
        for key,value in updates.items():
            updated[item_id][key] = value
    return updated
